@@ -1,5 +1,6 @@
 import pickle
 import random
+import numpy
 
 from fractions import Fraction as Q
 
@@ -71,6 +72,13 @@ def test_multiplication_by_one():
     assert a * 1 == a
 
 
+def test_multiplication_with_array():
+    p, k, n = 10007, 3, 2
+    a = PAdic(random.randrange(0, 10000), p, k, n)
+    array = [0, 1, 2, 3]
+    assert numpy.all(a * numpy.array(array) == numpy.array([a * entry for entry in array]))
+
+
 def test_unary_neg():
     p, k, n = 10007, 3, 2
     a = PAdic(random.randrange(0, 10000), p, k, n)
@@ -125,3 +133,7 @@ def test_padic_log():
     padic_log(w)  # branch info?
     w = PAdic(2, p, 5, +3)
     padic_log(w)  # branch info?
+
+
+def test_trivial_padic_log():
+    assert padic_log(1) == 0
